//      Device
//   

#include <conio.h>
#include <stdlib.h>
#include <iostream.h>
#include <fstream.h>
#include <stdio.h>
#include <string.h>
#include "in.cpp"

//===================== Globals ==============================//

class Node;		//    " "

enum SortBy		//  
{ 
	None, Name, Num 
};

char buff[256];	//  

typedef unsigned int uint;

void Mes(char *s) 	//  
{
	printf ("\n\t %s\n",s);		fflush(stdin);
}

uint Menu()			//  
{
	printf ("\n\n\t\t0 - Quit"
			"\n\t\t1 - Show"
			"\n\t\t2 - Add"
			"\n\t\t3 - Delete"
			"\n\t\t4 - Open"
			"\n\t\t5 - Save"
			"\n\t\t6 - Sort\n\n");
	return  getch() -'0';
}

SortBy SortMenu()	//   
{
	printf ("\n\t\tSorting\n"
			"\n\t\t1 - By Name"
			"\n\t\t2 - By Number"
			"\n\t\t0 - Don't Sort\n\n");
	return (SortBy)(getch()-'0');
}

uint DevMenu()	//   
{
	printf ("\n\n\t\t1 - Generator"
			"\n\t\t2 - Oscilloscope"
			"\n\t\t0 - Stop\n\n");
	return (uint)(getch()-'0');
}

//===================== Class Device ========================//

class Device				//    
{
	friend class Node;		// ,   
	friend class List;		// ,   
protected:					//  
	char *Name;				//  
	uint Num;				//  
public:						//  
	Device()				//  
	{
		Name=NULL;		Num=0;
	}
	
	Device (const Device& d)
	{ 
		Num = d.Num;
		Name=strcpy(new char[strlen(d.Name)+1],d.Name);
	}

	~Device()
	{						//   
		printf("\nDeleting: %s\n",Name);
		delete [] Name;
	}
	
	virtual void InKey()	//   
	{
		printf("\nDevice Name: ");	gets(buff);
		Name=strcpy(new char[strlen(buff)+1],buff);
		printf("ID Number: ");
		Num = (uint)in(1,100000);
	}
							//     
	virtual void InFile(ifstream& is)
	{
		is.getline(buff,255);
		Name = strcpy(new char[strlen(buff)+1],buff);
		is.getline(buff,255);
		Num = (uint)strtol (buff, 0, 0);
	}

	virtual char* GetName()=0;	//   
								//     
	virtual void OutScreen(uint n)
	{
		printf("\n%d. ID Number: %d\t%s",n,Num,Name); 
	}
								//     
	virtual void OutFile(ofstream& os)
	{
		char *p;
		for (p=Name; *p;  p++)
			os.put(*p);
		os << '\n' << Num << '\n';
	}
};

//===================== Derived Class Generator =====================//

class Generator : public Device		//  
{
	friend class Node;
protected:					//  
	uint Freq;				//   
public:						//  
	Generator() : Device()
	{
		Freq=0;
	}
	Generator(const Generator& d) : Device(d)
	{
		Freq = d.Freq;
	}
							//   -
	virtual void InKey()
	{
		Device::InKey();
		printf("Frequency: ");	gets(buff);
		Freq = (uint)strtol(buff,0,0);
	}
		
	virtual void InFile(ifstream& is)
	{
		is.getline(buff,255);
		Freq = (uint)strtol (buff, 0, 0);
		Device::InFile(is);
	}
							//    
	virtual char* GetName()
	{
		return "Generator";
	}

	virtual void OutScreen(uint n)
	{
		Device::OutScreen(n);
		printf("\tFrequency: %d",Freq);
	}
	
	virtual void OutFile(ofstream& os)
	{
		os << 'G' << ' ' << Freq << '\n';
		Device::OutFile(os);
	}
};

//===================== Derived Class Oscilloscope =====================//

class Oscilloscope : public Device
{
	friend class Node;
protected:					//  
	double Sens;			//  
public:
	Oscilloscope() : Device()
	{
		Sens=0.;
	}
	Oscilloscope (const Oscilloscope& d) : Device (d)
	{
		Sens = d.Sens;
	}
							//   -
	virtual void InKey()
	{
		Device::InKey();
		printf("Sensitivity: ");	gets(buff);
		Sens = strtod (buff,0);
	}

	virtual void InFile(ifstream& is)
	{
		is.getline(buff,255);
		Sens = strtod (buff,0);
		Device::InFile(is);	
	}
							//    
	virtual char* GetName()
	{
		return "Oscilloscope";
	}

	virtual void OutScreen(uint n)
	{
		Device::OutScreen(n);
		printf("\tSensitivity: %g",Sens);
	}
	
	virtual void OutFile(ofstream& os)
	{
		os << 'O' << ' ' << Sens << '\n';
		Device::OutFile(os);
	}
};

//===================== Class Node ========================//

class Node					// ,   
{
public:
	Device *pDev;			//   
	Node *pNext;			//   
							// 
	Node()
	{
		pDev = 0;  pNext = 0;
	}
	Node(Device *p, Node *n)
	{
		pDev = p;  pNext = n;
	}
	~Node () { delete pDev; }		// 
};

//===================== Class List =======================//

class List
{
private:
	Node *First;			//    
	uint Nodes;				//   
public:
	List()
	{
		First=0;		Nodes=0;
	}
	~List() { ClearList(); }
	void Mode();			//  
	void Insert();			// 
	void Show();			//   
	void Delete();			// 
	void ClearList();		// 
	void Write();			//   
	void Read();			//   
							//   
	bool Order (Node*, Node*, SortBy);
	void Sort();			// 
};

void List::Mode()
{
	puts("\n\t\t Linear List");
	while (1)
	{
		switch (Menu())		//     
		{
		case 0:	return;		//    
		case 1:	Show();		break;
		case 2:	Insert();	break;
		case 3:	Delete();	break;
		case 4:	Read();		break;
		case 5:	Write();	break;
		case 6:	Sort();		break;
		default: Mes("Please, enter an integer ranging 0...6");
		}
	}
}

void List::Insert()			//     
{
	printf("\n\tInsert: Press \"Enter\" to stop inserting\n");
	Device *p;
	while (1) 				// ,  
	{
		switch (DevMenu())
		{
		case 1: p = new Generator;		break;
		case 2: p = new Oscilloscope;	break;
		default: return;
		}
		puts(p->GetName());			//   
		p->InKey();					//      
		First = new Node(p,First);	//    
		Nodes++;
	}
}

void List::Show()					//     
{
	Node *p;						//    
	uint i;							//   
									//    
	for (p=First,i=1;  i<=Nodes; i++, p=p->pNext)
		p->pDev->OutScreen(i);		//   

	if (!Nodes)						//   , 
		Mes("List is empty");		//  
}

void List::Delete()
{
	Node *c,*p;						//       
	uint i, n;
									//        
	while (Nodes)
	{
		printf("\n\tDelete\t0...%d  (0 - to Quit): ",Nodes);
		if (!(n = (uint)in(0,Nodes)))
			break;					//     
									//  n- 
		for (c=First, p=0, i=1;  i<n;  i++, c=(p=c)->pNext)
			;						//   
									//  
		if (p)
			p->pNext = c->pNext;
		else
			First = c->pNext;
									//  
		delete c;	
		Nodes--;
	}
	Show();							//   
}

void List::ClearList()
{
	Node *pNext, *c;				//      
	for (c=First; c; c=pNext)
	{
		pNext=c->pNext;		delete c;
	}
	First = 0;  Nodes=0;			//  
	puts("\nThe list is destroyed");
}

bool List::Order (Node *p1, Node *p2, SortBy field)
{
	switch (field) 					//     
	{
	case Name:						//   
		return strcmp (p1->pDev->Name, p2->pDev->Name)<=0;
	case Num:
		return p1->pDev->Num <= p2->pDev->Num;
	default:
		return true;
	}
}

void List::Sort()
{
	bool Sorted=false;				//   
	Node *p0, *p1, *p2, *t;			//  
	uint i, last;					//   
	if (!Nodes)
		return;						//  

	SortBy field = SortMenu();		//   
	if (field==None || field>Num)
		return;
									//   
	for (last=Nodes; !Sorted; last--)
		for (Sorted=true, p0=0, p1=First, p2=First->pNext, i=1;
				i<last;
				p0=p1, p1=p2, p2=p2->pNext, i++)
			if (!Order (p1, p2, field))
			{						//   
				Sorted=false;
				if (p0)
					p0->pNext = p2;		//  
				else	
					First = p2;			//  
				p1->pNext=p2->pNext;	//   
				p2->pNext=p1;			// 
				t=p1; p1=p2; p2=t;		//  
			}
	Show();								//   
}

void List::Write()						//     
{
	Node *p;							//   
	uint i;								//  

	if (!Nodes)
	{
		Mes("List is empty"); return;
	}
										//  
	ofstream os ("DevList.dat");
	if (!os)
	{
		Mes("Can't open DevList.dat"); return;
	}
	os << Nodes << '\n';
										//   ,     
	for (p=First, i=0;  i<Nodes;  i++, p=p->pNext)
		p->pDev->OutFile(os);
	os.close();
	Mes("List has been stored");
}

void List::Read()
{
	Device *p;
	uint i;
	char type;
	
	ifstream is ("DevList.dat",ios::nocreate|ios::in);
	if (!is)
	{
		Mes("Can't open DevList.dat"); return;
	}
										//   
	if (Nodes)
		ClearList();
	Mes("Restoring from DevList.dat:");
										//   
	is.getline(buff,255);
	Nodes = (uint)strtol (buff, 0, 0);
										//      
	for (i=0; i<Nodes; i++)
	{
		is >> type;
		switch (type)					//    
		{								//   
		case 'G': p = new Generator;			break;
		case 'O': p = new Oscilloscope;	break;
		}
		p -> InFile(is);				//    
										//    
		First = new Node(p,First);
	}
	is.close();
	Show();
}

//===================== Global main function =========================//

void main()		//  
{
	List().Mode();
}
