//      Device
//   

#include <conio.h>
#include <stdlib.h>
#include <iostream.h>
#include <fstream.h>
#include <stdio.h>
#include <string.h>
#include "in.cpp"

//===================== Globals ==============================//

typedef unsigned int uint;

char buff[256];

void Mes (char *s) { printf ("\n\n\t%s\n",s);	fflush(stdin); }

uint Menu() {
	printf ("\n\n\t\t0 - Quit"
			"\n\t\t1 - Show"
    		"\n\t\t2 - Add"
			"\n\t\t3 - Delete"
			"\n\t\t4 - Open"
			"\n\t\t5 - Save\n\n");
	return  getch() -'0';
}

uint DevMenu() {
	printf ("\n\n\t\t1 - Generator"
			"\n\t\t2 - Oscilloscope"
			"\n\t\t0 - Stop\n\n");
	return (uint)(getch()-'0');
}

//===================== Class Device ========================//

class Device				//    
{
	friend class Node;		// ,   
	friend class Tree;		// ,   
protected:					//  
	char *Name;				//  
	uint Num;				//  
public:						//  
	Device()				//  
	{
		Name=NULL;		Num=0;
	}
	
	Device (const Device& d)
	{ 
		Num = d.Num;
		Name=strcpy(new char[strlen(d.Name)+1],d.Name);
	}

	~Device()
	{						//   
		printf("\nDeleting: %s\n",Name);
		delete [] Name;
	}
	
	virtual void InKey()	//   
	{
		printf("\nDevice Name: ");	gets(buff);
		Name=strcpy(new char[strlen(buff)+1],buff);
		printf("ID Number: ");
		Num = (uint)in(1,100000);
	}
							//     
	virtual void InFile(ifstream& is)
	{
		is.getline(buff,255);
		Name = strcpy(new char[strlen(buff)+1],buff);
		is.getline(buff,255);
		Num = (uint)strtol (buff, 0, 0);
	}

	virtual char* GetName()=0;	//   
								//     
	virtual void OutScreen()
	{
		printf("\nID Number: %d\t%s",Num,Name); 
	}
								//     
	virtual void OutFile(ofstream& os)
	{
		char *p;
		for (p=Name; *p;  p++)
			os.put(*p);
		os << '\n' << Num << '\n';
	}
};

//===================== Derived Class Generator =====================//

class Generator : public Device		//  
{
	friend class Node;
protected:					//  
	uint Freq;				//   
public:						//  
	Generator() : Device()
	{
		Freq=0;
	}
	Generator(const Generator& d) : Device(d)
	{
		Freq = d.Freq;
	}
							//   -
	virtual void InKey()
	{
		Device::InKey();
		printf("Frequency: ");	gets(buff);
		Freq = (uint)strtol(buff,0,0);
	}
		
	virtual void InFile(ifstream& is)
	{
		is.getline(buff,255);
		Freq = (uint)strtol (buff, 0, 0);
		Device::InFile(is);
	}
							//    
	virtual char* GetName()
	{
		return "Generator";
	}

	virtual void OutScreen()
	{
		Device::OutScreen();
		printf("\tFrequency: %d",Freq);
	}
	
	virtual void OutFile(ofstream& os)
	{
		os << 'G' << ' ' << Freq << '\n';
		Device::OutFile(os);
	}
};

//===================== Derived Class Oscilloscope =====================//

class Oscilloscope : public Device
{
	friend class Node;
protected:					//  
	double Sens;			//  
public:
	Oscilloscope() : Device()
	{
		Sens=0.;
	}
	Oscilloscope (const Oscilloscope& d) : Device (d)
	{
		Sens = d.Sens;
	}
							//   -
	virtual void InKey()
	{
		Device::InKey();
		printf("Sensitivity: ");	gets(buff);
		Sens = strtod (buff,0);
	}

	virtual void InFile(ifstream& is)
	{
		is.getline(buff,255);
		Sens = strtod (buff,0);
		Device::InFile(is);	
	}
							//    
	virtual char* GetName()
	{
		return "Oscilloscope";
	}

	virtual void OutScreen()
	{
		Device::OutScreen();
		printf("\tSensitivity: %g",Sens);
	}
	
	virtual void OutFile(ofstream& os)
	{
		os << 'O' << ' ' << Sens << '\n';
		Device::OutFile(os);
	}
};

//===================== Class Node ========================//

class Node
{
public:
	Device *pDev;					//    
	Node *left, *right;				//  ,   
	Node()
	{ 
		pDev=0;  left=right=0; 
	}
	Node(Device *p)
	{
		pDev = p;  left=right=0; 
	}
	~Node () { delete pDev; }
};

//===================== Class Tree ========================//

class Tree
{
	Node *Root;						//  
	uint Nodes;						//  
	Device **Key;
	int CurNum;
public:
	Tree ()
	{
		Root=NULL;  Nodes=0; CurNum=0;
	}
	void Mode();					//   
	void Insert();					//  
	void Delete();					//  
	void Show(Node *n=0, bool bFirst=true);		//  
	void Clear(Node *n=0, bool bFirst=true);	//  
									//      
	void ToArray (Node*n=0, bool bFirst=true);
	void Read();					//   
	void Write();					//   
	~Tree();
};

void Tree::Mode() {
	puts("\n\t\t Binary Tree");
	while (1)
	{
		switch (Menu())
		{
		case 0:	return;
		case 1:	Show();		break;	
		case 2:	Insert();	break;
		case 3:	Delete();	break;
		case 4:	Read();		break;
		case 5:	Write();	break;
		default: Mes ("Please, enter an integer ranging 0...5");
		}
	}
}

void Tree::Insert()					//   
{
	char branch;
	Device *d;  
	Node *p, *c;
	while (1)						//   
	{
		switch (DevMenu())			//  ?
		{
		case 0: return;				//  
		case 1: 
			d=new Generator;
			break;
		case 2:
			d=new Oscilloscope;
			break;
		default: continue;			// 
		}
		
		d->InKey();					//  
		
			//   
		for (c=Root, p=0;  c && c->pDev->Num != d->Num; )
			if ((p=c)->pDev->Num < d->Num)		//  
			{
				c=c->right;  branch='r';
			}
			else								//  
			{
				c=c->left;  branch='l';
			}
		
			//        ,  
		if (c)
		{
			Mes("Device is already in the Tree");
			delete d;							//  
			continue;
		}
		
			//       
		Node *n=new Node(d);
		if (!p)
			Root = n;					//  
		else							//  
			if (branch=='r')
				p->right = n;			// 
			else
				p->left  = n;			// 
		Nodes++;
	}
}

void Tree::Delete()			//  
{
	Node *p, *c, *r, *q, *s;
	uint n;
	
	while (Nodes)
	{
		Show();
		Mes("Delete ID Number\t (0 - to Quit): ");
		
			//   -    
		if (!(n = (uint)in(0,uint(~0)>>1)))
			break;
		
			//  ,      n
		for ( c=Root, p=0; c && c->pDev->Num != n; )
			if ((p=c)->pDev->Num < n)
				c=c->right;
			else
				c=c->left;
					
		if (!c)		//  
		{
			Mes("The Tree has no such device");
			continue;		//  
		}
			
			//---- .   
		if (c->left)						//   
			if (c->right)					//  
			{
					//   
				for (r=q=c->right;  s=q->left;  q=s)
					;						//  ,  
				q->left=c->left;			//    
			}
			else
				r=c->left;					//    -  
		else
			r=c->right;						//    -  
											//   -   NULL
		if (p)								//     
			if (c==p->left)
				p->left=r;					//   
			else
				p->right=r;					//   
		else
			Root=r;							//    
		delete c;							//  
		Nodes--;
	}
}

void Tree::Show (Node *c, bool bFirst)
{
	if (bFirst)
	{							//   
		c=Root;
		if (!c)  Mes("Tree is empty");
	}
	if (!c)
		return;					//  
	
	Show (c->left, 0);			//   
	
	c->pDev->OutScreen();		//   
	
	Show (c->right, 0);			//  
}

void Tree::Clear (Node *c, bool bFirst)
{
	if (bFirst) 				//   
	{
		c=Root;  Root=NULL;
	}
	if (!c)
		return;					//  
	
	Clear (c->left, 0);			//   
	
	Node *right=c->right;
	
	delete c;					//  
	
	Clear (right, 0);			//  
}

Tree::~Tree() { 
	Clear(); 
	Mes("The Tree has been deleted"); 
}

void Tree::ToArray (Node *c, bool bFirst)
{
	if (bFirst) 				//   
	{
		c=Root;		CurNum=0;
		if (!c)  Mes("The Tree is empty");
	}
	if (!c)
		return;					//  
	
	ToArray(c->left, 0);		//   
	
	Key[CurNum++] = c->pDev;	//   
	
	ToArray(c->right, 0);		//  
}

void Tree::Write()		//       
{
	if (!Nodes) 
	{ 
		Mes("The Tree is empty"); return; 
	}
	
	ofstream os("Tree.dat");	//  
	
	if (!os) 
	{ 
		Mes("Can't open Tree.dat"); return; 
	}
	
	os << Nodes << '\n';		//    
	
	Key = new Device*[Nodes];	//    
	ToArray();					//   Key[]
	
		//       
	bool *Ind = new bool[Nodes];
	memset (Ind,false,Nodes*sizeof(bool));

		//   
	for (uint i=0;  i<Nodes;  i++)
	{
		int k;								//   
		while (Ind[k = rand() % Nodes])
			;								//   
		Ind[k]=true;						//   
		Device *c = Key[k];					//    
		c->OutFile(os);						//    ()
	}
	
	delete [] Key; Key=0;					//  
	delete [] Ind;
	
	Mes("The Tree is stored in Tree.dat");
}

void Tree::Read()		//       
{
	Node *c, *p;
	Device *d;
	char type, branch;
	
	ifstream is("Tree.dat");
	
	if (!is)
	{
		Mes("Can't open Tree.dat"); return;
	}
	
	if (Root)
		Clear();		//   
	
	Mes("Restoring from DevList.dat:");
	is >> Nodes;		//    
	
	Mes("New nodes succession:\n");
	
	for (uint i=0; i<Nodes; i++)
	{
		is >> type;
		switch (type)
		{
		case 'G':
			d=new Generator;
			break;
		case 'O':
			d=new Oscilloscope;
			break;
		default: continue;		// 
		}
		
		d->InFile(is);			//  
		
		printf("%d ",d->Num);	//   
			//-----   
		for (c=Root, p=0; c && c->pDev->Num != d->Num; )
			if ((p=c)->pDev->Num < d->Num)	//  
			{
				c=c->right;  branch='r';
			}
			else							//  
			{
				c=c->left;  branch='l';
			}
			
		if (c)						//   
		{
			Mes("Device is already in the Tree");
			continue;				// .   
		}
		
		Node *n=new Node(d);		//   
									//   
		if (!p) Root = n;			//  
		else						//  
			if (branch=='r')
				p->right = n;		// 
			else	
				p->left  = n;		// 
	}
	puts("\n");		Show();
}

void main()
{
	Tree().Mode();
}
